@@ -1,8 +1,24 @@
 from meshagent.api.protocol import Protocol, ClientProtocol
+from meshagent.api.specs.service import ServiceTemplateSpec
 import json
 import asyncio
 import logging
-from typing import Optional, Callable, Dict, List, Any, Literal, Generic, TypeVar
+from meshagent.api.participant_token import ApiScope
+from pydantic import BaseModel, Field, JsonValue, ConfigDict
+from typing import (
+    Optional,
+    Callable,
+    Dict,
+    List,
+    Any,
+    Literal,
+    Generic,
+    TypeVar,
+    AsyncIterator,
+    Awaitable,
+)
+
+import base64
 
 from meshagent.api.messaging import Request
 from meshagent.api.runtime import runtime, RuntimeDocument
@@ -19,11 +35,48 @@ from meshagent.api.messaging import (
     FileResponse,
     Response,
 )
+from meshagent.api.oauth import OAuthClientConfig, ConnectorRef
 import uuid
 
 from datetime import datetime
 
 from abc import ABC, abstractmethod
+from dataclasses import dataclass
+
+
+def decode_records(records: list[dict]):
+    for r in records:
+        if isinstance(r, dict):
+            for k in r.keys():
+                v = r[k]
+                if isinstance(v, dict):
+                    encoding = v["encoding"]
+                    if encoding == "base64":
+                        r[k] = base64.b64decode(v["data"].encode())
+                    else:
+                        raise ValueError(f"Invalid encoding type {encoding}")
+
+    return records
+
+
+def encode_records(records: list[dict]):
+    transformed_records = []
+
+    for r in records:
+        c = {}
+        for k in r.keys():
+            v = r[k]
+            if isinstance(v, bytes):
+                c[k] = {
+                    "encoding": "base64",
+                    "data": base64.b64encode(v).decode(),
+                }
+            else:
+                c[k] = v
+
+        transformed_records.append(c)
+
+    return transformed_records
 
 
 logger = logging.getLogger("room_server_client")
@@ -31,7 +84,14 @@ logger.setLevel(logging.WARN)
 
 
 class RoomException(Exception):
-    pass
+    def __init__(self, message: str, *, status_code: int = 400):
+        self.status_code = status_code
+        super().__init__(message)
+
+
+class RoomAccessDeniedException(RoomException):
+    def __init__(message: str):
+        super().__init__(message, status_code=403)
 
 
 class Requirement(ABC):
@@ -53,6 +113,13 @@ class Requirement(ABC):
         pass
 
 
+class _MakeCallRequest(BaseModel):
+    url: str
+    arguments: dict
+    name: str
+    api: Optional[ApiScope] = None
+
+
 class RequiredToolkit(Requirement):
     # Require a toolkit to be present for this tool to execute, optionally a list of specific tools in the toolkit
     def __init__(self, *, name: str, tools: Optional[list["str"]] = None):
@@ -174,7 +241,14 @@ class _QueuedRoomMessage(RoomMessage):
 
 
 class RoomClient:
-    def __init__(self, *, protocol: ClientProtocol):
+    def __init__(
+        self,
+        *,
+        protocol: ClientProtocol,
+        oauth_token_request_handler: Optional[
+            Callable[["OAuthTokenRequest"], Awaitable]
+        ] = None,
+    ):
         self.protocol = protocol
         self.protocol.register_handler("room_ready", self._handle_ready)
         self.protocol.register_handler("room.status", self._handle_status)
@@ -195,6 +269,10 @@ class RoomClient:
         self.developer = DeveloperClient(room=self)
         self.queues = QueuesClient(room=self)
         self.database = DatabaseClient(room=self)
+        self.containers = ContainersClient(room=self)
+        self.secrets = SecretsClient(
+            room=self, oauth_token_request_handler=oauth_token_request_handler
+        )
 
         self._room_url = None
         self._room_name = None
@@ -214,13 +292,27 @@ class RoomClient:
     async def __aenter__(self):
         await self.protocol.__aenter__()
 
-        await self._ready
+        async def startup():
+            await self._ready
 
-        await self.sync.start()
+            await self.sync.start()
 
-        await self.messaging.start()
+            await self.messaging.start()
+            await self._local_participant_ready
 
-        await self._local_participant_ready
+        async def closed():
+            # protect against early termination
+            await self.protocol.wait_for_close()
+
+        startup_task = asyncio.create_task(startup())
+        close_task = asyncio.create_task(closed())
+        _, pending = await asyncio.wait(
+            [
+                startup_task,
+                close_task,
+            ],
+            return_when=asyncio.FIRST_COMPLETED,
+        )
 
         return self
 
@@ -256,8 +348,8 @@ class RoomClient:
     async def send_request(
         self, type: str, request: dict, data: bytes | None = None
     ) -> FileResponse | None | dict | str:
-        logger.info("sending request %s", type)
         request_id = self.protocol.next_message_id()
+        logger.debug("sending request %s %s", request_id, type)
 
         pr = _PendingRequest()
         self._pending_requests[request_id] = pr
@@ -265,7 +357,10 @@ class RoomClient:
         message = pack_message(header=request, data=data)
 
         await self.protocol.send(type=type, data=message, message_id=request_id)
-        return await pr.fut
+        result = await pr.fut
+        logger.debug("returning response %s", type)
+
+        return result
 
     async def _handle_status(
         self, protocol: Protocol, message_id: int, type: str, data: bytes
@@ -288,8 +383,6 @@ class RoomClient:
     async def _handle_response(
         self, protocol: Protocol, message_id: int, type: str, data: bytes
     ) -> None:
-        logger.info("received response %s", type)
-
         response = unpack_response(data=data)
 
         request_id = message_id
@@ -300,7 +393,7 @@ class RoomClient:
             else:
                 pr.fut.set_result(response)
         else:
-            logger.warning(
+            logger.error(
                 "received a response for a request that is not pending {id}".format(
                     id=request_id
                 )
@@ -371,7 +464,8 @@ class SyncClient:
     async def stop(self):
         self._sync_ch.close()
 
-        asyncio.gather(self._main_task)
+        if self._main_task is not None:
+            asyncio.gather(self._main_task)
 
     async def create(self, *, path: str, json: Optional[dict] = None) -> None:
         await self.room.send_request("room.create", {"path": path, "json": json})
@@ -415,6 +509,10 @@ class SyncClient:
             self._connecting_documents.pop(path)
 
             logger.info("Connected to %s", path)
+
+        except asyncio.CancelledError:
+            raise
+
         except Exception as e:
             connecting_fut.set_exception(e)
             self._connecting_documents.pop(path)
@@ -458,9 +556,7 @@ class SyncClient:
             if not doc.ref.synchronized.done():
                 doc.ref.synchronized.set_result(True)
         else:
-            raise RoomException(
-                "received change for a document that is not connected:" + path
-            )
+            logger.debug("received change for a document that is not connected:" + path)
 
 
 class AgentDescription:
@@ -502,7 +598,7 @@ class ToolDescription:
         thumbnail_url: Optional[str] = None,
         defs: Optional[dict] = None,
         pricing: Optional[str] = None,
-        supports_context: Optional[bool] = None,
+        supports_context: Optional[bool] = False,
     ):
         self.name = name
         self.title = title
@@ -565,9 +661,14 @@ class AgentsClient:
     def __init__(self, *, room: RoomClient):
         self.room = room
 
-    async def make_call(self, *, name: str, url: str, arguments: dict) -> None:
+    async def make_call(
+        self, *, name: str, url: str, arguments: dict, api: Optional[ApiScope] = None
+    ) -> None:
         await self.room.send_request(
-            "agent.call", {"name": name, "url": url, "arguments": arguments}
+            "agent.call",
+            _MakeCallRequest(
+                name=name, url=url, arguments=arguments, api=api
+            ).model_dump(mode="json"),
         )
         return None
 
@@ -738,7 +839,11 @@ class LivekitClient:
 
 class StorageEntry:
     def __init__(
-        self, name: str, is_folder: bool, created_at: datetime, updated_at: datetime
+        self,
+        name: str,
+        is_folder: bool,
+        created_at: Optional[datetime] = None,
+        updated_at: Optional[datetime] = None,
     ):
         self.name = name
         self.is_folder = is_folder
@@ -1055,6 +1160,13 @@ class MessagingClient:
     def get_participants(self) -> list[RemoteParticipant]:
         return list(self._participants.values())
 
+    def get_participant(self, id: str) -> RemoteParticipant | None:
+        for part in self.remote_participants:
+            if part.id == id:
+                return part
+
+        return None
+
     async def enable(
         self,
         *,
@@ -1104,7 +1216,8 @@ class MessagingClient:
 
     async def stop(self):
         self._message_queue.close()
-        await asyncio.gather(self._send_task)
+        if self._send_task is not None:
+            await asyncio.gather(self._send_task)
 
     async def _send_messages(self):
         async for msg in self._message_queue:
@@ -1120,6 +1233,9 @@ class MessagingClient:
                 )
                 msg.fut.set_result(True)
 
+            except asyncio.CancelledError:
+                raise
+
             except Exception as ex:
                 logger.info("Unable to send message to participant", exc_info=ex)
                 msg.fut.set_exception(ex)
@@ -1244,6 +1360,7 @@ class MessagingClient:
         def on_send_complete(task: asyncio.Task):
             try:
                 task.result()
+
             except Exception as e:
                 logger.warning("unable to send stream response", exc_info=e)
 
@@ -1290,6 +1407,9 @@ class MessagingClient:
             )
             send.add_done_callback(on_send_complete)
 
+        except asyncio.CancelledError:
+            raise
+
         except Exception as e:
             logger.info(f"rejecting stream {stream_id}")
             # Reject
@@ -1596,6 +1716,21 @@ class TextDataType(DataType):
 _data_types["text"] = TextDataType
 
 
+class BinaryDataType(DataType):
+    def __init__(self):
+        super().__init__()
+
+    @staticmethod
+    def from_json(data: dict):
+        assert data["type"] == "binary"
+        return BinaryDataType()
+
+    def to_json(self):
+        return {"type": "binary"}
+
+
+_data_types["binary"] = BinaryDataType
+
 CreateMode = Literal["create", "overwrite", "create_if_not_exists"]
 
 
@@ -1669,19 +1804,31 @@ class DatabaseClient:
     ) -> None:
         return await self._create_table(name=name, data=data, mode=mode)
 
-    async def drop_table(self, *, name: str, ignore_missing: bool = False):
+    async def drop_table(self, *, name: str, ignore_missing: bool = False) -> None:
         """
         Drop (delete) a table.
 
         :param name: Table name.
         :param ignore_missing: If True, ignore if table doesn't exist.
-        :return: Server response dict containing "status", "table", etc.
         """
         payload = {"name": name, "ignore_missing": ignore_missing}
         await self.room.send_request("database.drop_table", payload)
         return None
 
-    async def add_columns(self, *, table: str, new_columns: Dict[str, str]) -> None:
+    async def drop_index(self, *, table: str, name: str) -> None:
+        """
+        Drop (delete) a index.
+
+        :param table: table name
+        :param name: index name.
+        """
+        payload = {"table": table, "name": name}
+        await self.room.send_request("database.drop_index", payload)
+        return None
+
+    async def add_columns(
+        self, *, table: str, new_columns: Dict[str, str | DataType]
+    ) -> None:
         """
         Add new columns to an existing table.
 
@@ -1689,7 +1836,16 @@ class DatabaseClient:
         :param new_columns: Dict of {column_name: default_value_expression}.
         """
 
-        payload = {"table": table, "new_columns": new_columns}
+        columns = {}
+
+        for c in new_columns.keys():
+            if isinstance(new_columns[c], DataType):
+                columns[c] = new_columns[c].to_json()
+            else:
+                columns[c] = new_columns[c]
+
+        payload = {"table": table, "new_columns": columns}
+
         await self.room.send_request("database.add_columns", payload)
         return None
 
@@ -1699,7 +1855,6 @@ class DatabaseClient:
 
         :param table: Table name.
         :param columns: List of column names to drop.
-        :return: Server response dict with "status", "table", "dropped_columns".
         """
         payload = {"table": table, "columns": columns}
 
@@ -1712,11 +1867,11 @@ class DatabaseClient:
 
         :param table: Table name.
         :param records: The record(s) to insert (list or dict).
-        :return: Server response dict with "status", "table", "result".
         """
+
         payload = {
             "table": table,
-            "records": records,
+            "records": encode_records(records),
         }
         await self.room.send_request("database.insert", payload)
 
@@ -1735,7 +1890,6 @@ class DatabaseClient:
         :param where: SQL WHERE clause (e.g. "id = 123").
         :param values: Dict of column updates, e.g. {"col1": "new_value"}.
         :param values_sql: Dict of SQL expressions for updates, e.g. {"col2": "col2 + 1"}.
-        :return: Server response dict with "status", "table", "where".
         """
         payload = {
             "table": table,
@@ -1751,7 +1905,6 @@ class DatabaseClient:
 
         :param table: Table name.
         :param where: SQL WHERE clause (e.g. "id = 123").
-        :return: Server response dict with "status", "table", "where".
         """
         payload = {"table": table, "where": where}
         await self.room.send_request("database.delete", payload)
@@ -1765,7 +1918,6 @@ class DatabaseClient:
         :param table: Table name.
         :param on: Column name to match on (e.g. "id").
         :param records: The record(s) to merge.
-        :return: Server response dict with "status", "table", "on".
         """
         payload = {"table": table, "on": on, "records": records}
         await self.room.send_request("database.merge", payload)
@@ -1778,6 +1930,7 @@ class DatabaseClient:
         text: Optional[str] = None,
         vector: Optional[list[float]] = None,
         where: Optional[str] | dict = None,
+        offset: Optional[int] = None,
         limit: Optional[int] = None,
         select: Optional[List[str]] = None,
     ) -> list[Dict[str, Any]]:
@@ -1789,7 +1942,6 @@ class DatabaseClient:
         :param where: A filter clause or values to match
         :param limit: Limit the number of results.
         :param select: Columns to select.
-        :return: Server response dict with "status", "table", "results".
         """
 
         if isinstance(where, dict):
@@ -1803,6 +1955,10 @@ class DatabaseClient:
         }
         if limit is not None:
             payload["limit"] = limit
+
+        if offset is not None:
+            payload["offset"] = offset
+
         if select is not None:
             payload["select"] = select
 
@@ -1810,11 +1966,11 @@ class DatabaseClient:
             payload["vector"] = vector
 
         response = await self.room.send_request("database.search", payload)
-        if hasattr(response, "json"):
-            return response.json["results"]
+        if isinstance(response, JsonResponse):
+            return decode_records(response.json["results"])
         return []
 
-    async def optimize(self, table: str) -> None:
+    async def optimize(self, *, table: str) -> None:
         """
         Optimize (compact/prune) a table.
 
@@ -1826,6 +1982,44 @@ class DatabaseClient:
         await self.room.send_request("database.optimize", payload)
         return None
 
+    async def restore(self, *, table: str, version: int) -> None:
+        """
+        restore a table version.
+
+        :param table: Table name.
+        """
+        payload = {
+            "table": table,
+            "version": version,
+        }
+        await self.room.send_request("database.restore", payload)
+        return None
+
+    async def checkout(self, *, table: str, version: int) -> None:
+        """
+        checkout a table version.
+
+        :param table: Table name.
+        """
+        payload = {
+            "table": table,
+            "version": version,
+        }
+        await self.room.send_request("database.checkout", payload)
+        return None
+
+    async def list_versions(self, *, table: str) -> list["TableVersion"]:
+        """
+        list a table's versions
+
+        :param table: Table name.
+        """
+        payload = {
+            "table": table,
+        }
+        resp = await self.room.send_request("database.list_versions", payload)
+        return [TableVersion.model_validate(v) for v in resp.json["versions"]]
+
     async def create_vector_index(
         self, *, table: str, column: str, replace: Optional[bool] = None
     ) -> None:
@@ -1877,7 +2071,7 @@ class DatabaseClient:
         await self.room.send_request("database.create_full_text_search_index", payload)
         return None
 
-    async def list_indexes(self, *, table: str) -> Dict[str, Any]:
+    async def list_indexes(self, *, table: str) -> list["TableIndex"]:
         """
         List all indexes on a table.
 
@@ -1886,6 +2080,724 @@ class DatabaseClient:
         payload = {"table": table}
         response = await self.room.send_request("database.list_indexes", payload)
         if hasattr(response, "json"):
-            return response.json["indexes"]
+            return [TableIndex.model_validate(i) for i in response.json["indexes"]]
 
         raise RoomException("unexpected return type")
+
+
+class TableVersion(BaseModel):
+    timestamp: datetime
+    version: int
+    metadata: dict[str, JsonValue]
+
+
+class TableIndex(BaseModel):
+    name: str
+    columns: list[str]
+    type: str
+
+
+class ProgressDetail(BaseModel):
+    current: Optional[int] = None
+    total: Optional[int] = None
+
+
+class LogProgress(BaseModel):
+    layer: Optional[str] = None
+    message: Optional[str] = None
+    current: Optional[int] = None
+    total: Optional[int] = None
+
+
+class ErrorDetail(BaseModel):
+    """Structured error information returned on failure."""
+
+    code: Optional[int] = None
+    message: str
+
+
+class PullMessage(BaseModel):
+    """
+    One JSON object emitted by the Engine while pulling *or* pushing.
+
+    Docker can add extra keys in new versions, so we allow unknown fields.
+    """
+
+    # Main variants ----------------------------------------------------------
+    status: Optional[str] = None  # layer status ("Downloading", "Extracting", ...)
+    id: Optional[str] = Field(None, alias="id")  # layer identifier / step number
+
+    # Progress bar -----------------------------------------------------------
+    progress: Optional[str] = None
+    progress_detail: Optional[ProgressDetail] = Field(None, alias="progressDetail")
+
+    # Success / aux payload --------------------------------------------------
+    aux: Optional[Any] = None  # e.g. {"Digest": "sha256:…", "Size": 123}
+
+    # Error handling ---------------------------------------------------------
+    error: Optional[str] = None
+    error_detail: Optional[ErrorDetail] = Field(None, alias="errorDetail")
+
+    # Misc extras sometimes present -----------------------------------------
+    time: Optional[int] = None  # seconds since epoch
+    from_: Optional[str] = Field(None, alias="from")  # reserve‑word workaround
+
+    model_config = ConfigDict(
+        validate_by_name=True,  # accept field aliases on input
+        extra="allow",  # keep unknown keys for forward‑compat
+    )
+
+
+class Image(BaseModel):
+    id: str
+    tags: List[str]
+    size: int
+    labels: Dict[str, str]
+    manifest: Optional[Any] = None
+
+
+class DockerSecret(BaseModel):
+    registry: Optional[str] = None
+    username: str
+    password: str
+
+
+class ImagePullRequest(BaseModel):
+    tag: str
+    credentials: List[DockerSecret] = Field(default_factory=list)
+
+
+class ListContainersRequest(BaseModel):
+    all: Optional[bool] = None
+
+
+class RunRequest(BaseModel):
+    request_id: Optional[str] = None
+    image: str
+    command: Optional[str] = None
+    env: Dict[str, str] = Field(default_factory=dict)
+    mount_path: Optional[str] = None
+    mount_subpath: Optional[str] = None
+    role: Optional[str] = None
+    participant_name: Optional[str] = None
+    ports: Dict[int, int] = Field(default_factory=dict)
+    credentials: Optional[List[DockerSecret]] = None
+    variables: Optional[Dict[str, str]] = None
+    name: Optional[str] = None
+    annotations: Optional[Dict[str, str]] = None
+
+
+class ExecRequest(BaseModel):
+    request_id: Optional[str] = None
+    container_id: str
+    command: Optional[str] = None
+    tty: Optional[bool] = None
+    detach: bool = True
+
+
+class ContainerRunResult(BaseModel):
+    container_id: str
+    status: Optional[int] = None
+    logs: List[str] = Field(default_factory=list)
+
+
+class ContainerStartedBy(BaseModel):
+    id: str
+    name: str
+
+
+class RoomContainer(BaseModel):
+    id: str
+    image: Optional[str] = None
+    status: Optional[str] = None
+    name: Optional[str] = None
+    started_by: ContainerStartedBy
+    manifest: Optional[ServiceTemplateSpec] = None
+    state: Literal["CREATED", "RUNNING", "EXITED", "UNKNOWN"]
+
+    # Accept arbitrary extras (names, created, state, etc.)
+    class Config:
+        extra = "allow"
+
+
+# ---------------------------
+# LogStream (awaitable + async generators)
+# ---------------------------
+
+T = TypeVar("T")
+
+
+class LogStream(Generic[T]):
+    """
+    - await stream: waits for final result (T or None)
+    - stream.logs(): async iterator of text lines
+    - stream.progress(): async iterator of LogProgress
+    - stream.cancel(): cancels on server
+    """
+
+    def __init__(
+        self,
+        *,
+        task: asyncio.Task,
+        cancel_cb: Callable[[], asyncio.Future[Any]],
+    ):
+        self._logs_q = asyncio.Queue[Optional[str]]()
+        self._progress_q = asyncio.Queue[Optional[LogProgress]]()
+
+        self._cancel_cb = cancel_cb
+        self._task = task
+
+        task.add_done_callback(self._on_task_done)
+
+    def _on_task_done(self, t):
+        self._logs_q.put_nowait(None)
+        self._progress_q.put_nowait(None)
+
+    @property
+    def result(self):
+        return asyncio.ensure_future(self._task)
+
+    def __await__(self):
+        return self.result.__await__()
+
+    async def cancel(self):
+        await self._cancel_cb()
+
+    async def logs(self) -> AsyncIterator[str]:
+        while True:
+            line = await self._logs_q.get()
+            if line is None:  # sentinel
+                return
+            yield line
+
+    async def progress(self) -> AsyncIterator[LogProgress]:
+        while True:
+            p = await self._progress_q.get()
+            if p is None:  # sentinel
+                return
+            yield p
+
+
+# ---------------------------
+# Container TTY
+# ---------------------------
+
+
+class Container:
+    """
+    Provides async input/output streams for an interactive container session.
+    """
+
+    def __init__(self, *, room: RoomClient, request_id: str, task: asyncio.Task):
+        self._room = room
+        self._request_id = request_id
+        self._output_q: asyncio.Queue[Optional[bytes]] = asyncio.Queue()
+        self._closed = asyncio.ensure_future(task)
+        self._task = task
+
+        task.add_done_callback(self._on_task_done)
+
+    def _on_task_done(self, t):
+        self._output_q.put_nowait(None)
+
+    @property
+    def result(self):
+        return self._closed
+
+    @property
+    def request_id(self):
+        return self._request_id
+
+    async def write(self, data: bytes) -> None:
+        # If server supports TTY input; adjust route name if different.
+        await self._room.send_request(
+            "containers.container_input",
+            {"request_id": self._request_id, "channel": 1},
+            data=data,
+        )
+
+    async def resize(self, *, width: int, height: int) -> None:
+        """
+        Resize the TTY for the running container.
+        This sends a control message (channel 4) to adjust terminal dimensions.
+        """
+        await self._room.send_request(
+            "containers.container_input",
+            {
+                "request_id": self._request_id,
+                "channel": 4,
+                "width": width,
+                "height": height,
+            },
+        )
+
+    async def output(self) -> AsyncIterator[bytes]:
+        while True:
+            chunk = await self._output_q.get()
+            if chunk is None:
+                return
+            yield chunk
+
+    async def kill(self):
+        # send a kill message on channel 5
+        await self._room.send_request(
+            "containers.container_input",
+            {"request_id": self._request_id, "channel": 5},
+            data={},
+        )
+
+    # Internal
+    def _push_output(self, data: bytes):
+        self._output_q.put_nowait(data)
+
+
+# ---------------------------
+# ContainersClient
+# ---------------------------
+
+
+class ContainersClient:
+    def __init__(self, *, room: RoomClient):
+        self.room = room
+        # Hook server -> client events
+        self.room.protocol.register_handler(
+            "containers.log.chunk", self._handle_log_chunk
+        )
+        self.room.protocol.register_handler(
+            "containers.run.output", self._handle_container_run_chunk
+        )
+        self.room.protocol.register_handler(
+            "containers.progress", self._handle_progress
+        )
+
+        self._ttys: Dict[str, Container] = {}
+        self._log_streams = dict[str, LogStream]()
+
+    # ---- Event handlers ----
+
+    async def _handle_log_chunk(
+        self, protocol: Protocol, message_id: int, typ: str, data: bytes
+    ):
+        header, _ = unpack_message(data)
+        req_id = header["request_id"]
+        log_line = header.get("log", "")
+        q = self._log_streams.get(req_id)
+        if q:
+            q._logs_q.put_nowait(str(log_line))
+
+    async def _handle_container_run_chunk(
+        self, protocol: Protocol, message_id: int, typ: str, data: bytes
+    ):
+        header, payload = unpack_message(data)
+
+        req_id: str = header["request_id"]
+        channel: int = int(header["channel"])
+        tty = self._ttys.get(req_id)
+        if tty is None:
+            logger.warning("received output from missing container %s", req_id)
+            return  # tty closed or missing
+
+        if channel == 0:
+            tty._push_output(payload)
+
+    async def _handle_progress(
+        self, protocol: Protocol, message_id: int, typ: str, data: bytes
+    ):
+        header, _ = unpack_message(data)
+        req_id = header["request_id"]
+        detail = header.get("detail") or {}
+        lp = LogProgress(
+            layer=header.get("layer"),
+            message=header.get("message"),
+            current=(detail.get("current") if detail else None),
+            total=(detail.get("total") if detail else None),
+        )
+        pq = self._log_streams.get(req_id)
+        if pq:
+            pq._progress_q.put_nowait(lp)
+
+    # ---- High-level API ----
+
+    async def list_images(self) -> List[Image]:
+        res = await self.room.send_request("containers.list_images", {})
+        imgs = res["images"]
+        return [Image.model_validate(i) for i in imgs]
+
+    async def delete_image(self, *, image: str) -> None:
+        await self.room.send_request("containers.delete_image", {"image": image})
+
+    # ---- Streaming helpers ----
+
+    def _make_stream(
+        self,
+        *,
+        cancel_cb: Callable[[], asyncio.Future[Any]],
+        task: asyncio.Task,
+        request_id: str,
+    ) -> LogStream:
+        log_stream = LogStream(cancel_cb=cancel_cb, task=task)
+        self._log_streams[request_id] = log_stream
+
+        def _pop(t):
+            self._log_streams.pop(request_id)
+
+        task.add_done_callback(_pop)
+
+        return log_stream
+
+    async def pull_image(
+        self, *, tag: str, credentials: List[DockerSecret] | None = None
+    ) -> None:
+        req = ImagePullRequest(tag=tag, credentials=credentials or [])
+
+        await self.room.send_request("containers.pull_image", req.model_dump())
+
+        return None
+
+    # ---- Run Container ----
+
+    async def run(
+        self,
+        *,
+        image: str,
+        command: Optional[str] = None,
+        env: Dict[str, str] | None = None,
+        mount_path: Optional[str] = None,
+        mount_subpath: Optional[str] = None,
+        role: Optional[str] = None,
+        participant_name: Optional[str] = None,
+        ports: Dict[int, int] | None = None,
+        variables: Optional[Dict[str, str]] = None,
+        credentials: List[DockerSecret] | None = None,
+        name: Optional[str] = None,
+    ) -> str:
+        request_id = uuid.uuid4().hex
+
+        req = RunRequest(
+            name=name,
+            request_id=request_id,
+            image=image,
+            command=command,
+            env=env or {},
+            mount_path=mount_path,
+            mount_subpath=mount_subpath,
+            role=role,
+            participant_name=participant_name,
+            ports=ports or {},
+            credentials=credentials or [],
+            variables=variables,
+        )
+
+        resp = await self.room.send_request(
+            "containers.run", req.model_dump(exclude_none=True)
+        )
+        if isinstance(resp, JsonResponse):
+            container_id: str = resp.json["container_id"]
+            return container_id
+
+        else:
+            raise RoomException(f"Unexpected response type {resp}")
+
+    async def exec(
+        self,
+        *,
+        container_id: str,
+        command: Optional[list[str]] = None,
+        tty: Optional[bool] = None,
+        detach: bool = True,
+    ) -> Container:
+        request_id = str(uuid.uuid4())
+
+        req = ExecRequest(
+            request_id=request_id,
+            container_id=container_id,
+            command=command,
+            detach=detach,
+            tty=tty,
+        )
+
+        async def run():
+            try:
+                resp = await self.room.send_request(
+                    "containers.exec", req.model_dump(exclude_none=True)
+                )
+
+                # close TTY on completion
+
+                status = (
+                    resp.json["status"]
+                    if isinstance(resp, JsonResponse)
+                    else resp.get("status", "")
+                )
+                return status
+            finally:
+                self._ttys.pop(request_id, None)
+
+        container = Container(
+            room=self.room, request_id=request_id, task=asyncio.create_task(run())
+        )
+        self._ttys[request_id] = container
+
+        return container
+
+    # ---- Logs ----
+
+    def logs(self, *, container_id: str, follow: bool = False) -> LogStream[None]:
+        request_id = uuid.uuid4().hex
+
+        async def cancel():
+            await self.room.send_request(
+                "containers.stop_logs", {"request_id": request_id}
+            )
+
+        async def _run():
+            await self.room.send_request(
+                "containers.logs",
+                {"request_id": request_id, "id": container_id, "follow": follow},
+            )
+            return None
+
+        stream = self._make_stream(
+            cancel_cb=cancel, task=asyncio.create_task(_run()), request_id=request_id
+        )
+        return stream
+
+    # ---- Misc ----
+
+    async def stop(self, *, container_id: str, force: bool = False) -> None:
+        await self.room.send_request(
+            "containers.stop_container", {"id": container_id, "force": force}
+        )
+
+    async def delete(self, *, container_id: str) -> None:
+        await self.room.send_request(
+            "containers.delete_container", {"id": container_id}
+        )
+
+    async def list(self, all: Optional[bool] = None) -> List[RoomContainer]:
+        res = await self.room.send_request(
+            "containers.list_containers",
+            ListContainersRequest(all=all).model_dump(mode="json"),
+        )
+        return [RoomContainer(**c) for c in res["containers"]]
+
+
+class _GetOfflineOAuthTokenRequest(BaseModel):
+    connector: Optional[ConnectorRef] = None
+    oauth: Optional[OAuthClientConfig] = None
+    delegated_to: Optional[str] = None
+    delegated_by: Optional[str] = None
+
+
+class _GetOfflineOAuthTokenResponse(BaseModel):
+    access_token: Optional[str] = None
+
+
+class _RequestOAuthTokenRequest(BaseModel):
+    connector: Optional[ConnectorRef] = None
+    oauth: Optional[OAuthClientConfig] = None
+    redirect_uri: str
+    participant_id: str
+    timeout: int = 60 * 5
+    redirect_uri: str
+    delegate_to: Optional[str] = None
+
+
+class _RequestOAuthTokenResponse(BaseModel):
+    access_token: Optional[str] = None
+
+
+class _DeleteUserSecretRequest(BaseModel):
+    id: str
+    delegated_to: Optional[str] = None
+
+
+class _DeleteUserSecretResponse(BaseModel):
+    pass
+
+
+class _ListUserSecretsRequest(BaseModel):
+    pass
+
+
+class SecretInfo(BaseModel):
+    id: str
+    type: str
+    name: str
+    delegated_to: Optional[str] = None
+
+
+class _ListUserSecretsResponse(BaseModel):
+    secrets: list[SecretInfo]
+
+
+class OAuthCredentials(BaseModel):
+    access_token: str
+    refresh_token: Optional[str] = None
+    expiration: Optional[datetime] = None
+    scopes: Optional[list[str]] = None
+
+
+class _ClientRequestOAuthTokenRequest(BaseModel):
+    request_id: str
+    request: _RequestOAuthTokenRequest
+    challenge: Optional[str]
+
+
+class _ClientRequestOAuthTokenResponse(BaseModel):
+    request_id: str
+    code: Optional[str] = None
+    error: Optional[str] = None
+
+
+@dataclass
+class OAuthTokenRequest:
+    request_id: str
+    authorization_endpoint: str
+    token_endpoint: str
+    challenge: str
+    scopes: Optional[list[str]] = None
+
+
+class SecretsClient:
+    def __init__(
+        self,
+        *,
+        room: RoomClient,
+        oauth_token_request_handler: Optional[
+            Callable[[OAuthTokenRequest], Awaitable]
+        ] = None,
+    ):
+        self.room = room
+        # Hook server -> client events
+        self.room.protocol.register_handler(
+            "secrets.request_oauth_token", self._handle_client_oauth_token_request
+        )
+        self._oauth_token_request_handler = oauth_token_request_handler
+        self._pending_authorization_requests = []
+
+    async def _handle_client_oauth_token_request(
+        self, protocol: Protocol, message_id: int, type: str, data: bytes
+    ) -> None:
+        request, bytes = unpack_message(data=data)
+        req = _ClientRequestOAuthTokenRequest.model_validate(request)
+
+        if self._oauth_token_request_handler is None:
+            raise RoomException("No oauth token handler registered")
+
+        def on_done(t: asyncio.Task):
+            try:
+                t.result()
+            finally:
+                self._pending_authorization_requests.remove(t)
+
+        task = asyncio.create_task(
+            self._oauth_token_request_handler(
+                OAuthTokenRequest(
+                    request_id=req.request_id,
+                    authorization_endpoint=req.request.oauth.authorization_endpoint,
+                    token_endpoint=req.request.oauth.token_endpoint,
+                    scopes=req.request.oauth.scopes,
+                    challenge=req.challenge,
+                )
+            )
+        )
+        task.add_done_callback(on_done)
+        self._pending_authorization_requests.append(task)
+
+    async def provide_oauth_authorization(
+        self,
+        *,
+        request_id: str,
+        code: str,
+    ):
+        await self.room.send_request(
+            "secrets.provide_oauth_authorization",
+            {
+                "code": code,
+                "request_id": request_id,
+            },
+        )
+
+    async def reject_oauth_authorization(
+        self,
+        *,
+        request_id: str,
+        error: str,
+    ):
+        await self.room.send_request(
+            "secrets.provide_oauth_authorization",
+            {
+                "error": error,
+                "request_id": request_id,
+            },
+        )
+
+    # get a saved oauth token
+    async def get_offline_oauth_token(
+        self,
+        *,
+        connector: Optional[ConnectorRef] = None,
+        oauth: Optional[OAuthClientConfig] = None,
+        delegated_to: Optional[str] = None,
+        delegated_by: Optional[str] = None,
+    ):
+        req = _GetOfflineOAuthTokenRequest(
+            connector=connector,
+            oauth=oauth,
+            delegated_by=delegated_by,
+            delegated_to=delegated_to,
+        )
+        response = await self.room.send_request(
+            "secrets.get_offline_oauth_token", req.model_dump(mode="json")
+        )
+        if isinstance(response, JsonResponse):
+            resp = _GetOfflineOAuthTokenResponse.model_validate(response.json)
+            return resp.access_token
+        else:
+            raise RoomException("Invalid response received, expected JsonResponse")
+
+    async def request_oauth_token(
+        self,
+        *,
+        connector: Optional[ConnectorRef] = None,
+        oauth: Optional[OAuthClientConfig] = None,
+        timeout: int = 60 * 5,
+        from_participant_id: str,
+        redirect_uri: str,
+        delegate_to: Optional[str] = None,
+    ) -> str | None:
+        req = _RequestOAuthTokenRequest(
+            redirect_uri=redirect_uri,
+            timeout=timeout,
+            participant_id=from_participant_id,
+            oauth=oauth,
+            connector=connector,
+            delegate_to=delegate_to,
+        )
+        response = await self.room.send_request(
+            "secrets.request_oauth_token", req.model_dump(mode="json")
+        )
+        if isinstance(response, JsonResponse):
+            resp = _RequestOAuthTokenResponse.model_validate(response.json)
+            return resp.access_token
+        else:
+            raise RoomException("Invalid response received, expected JsonResponse")
+
+    async def list_user_secrets(self) -> list[SecretInfo]:
+        response = await self.room.send_request(
+            "secrets.list_secrets", _ListUserSecretsRequest().model_dump(mode="json")
+        )
+        if isinstance(response, JsonResponse):
+            resp = _ListUserSecretsResponse.model_validate(response.json)
+            return resp.secrets
+        else:
+            raise RoomException("Invalid response received, expected JsonResponse")
+
+    async def delete_user_secret(self, *, id: str, delegated_to: Optional[str] = None):
+        await self.room.send_request(
+            "secrets.delete_secret",
+            _DeleteUserSecretRequest(id=id, delegated_to=delegated_to).model_dump(
+                mode="json"
+            ),
+        )
